/**
 * AutoDAO - Generic DAO on steroids implementation for Java.
 * Copyright (C) 2008-2009  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.autodao;

import java.io.Serializable;

import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.springframework.dao.DataAccessException;

/**
 * Generic interface for all DAOs.
 * @param <E> entity.
 * @param <K> primary key.
 */
public interface Dao<E extends PersistentEntity<K>, K extends Serializable> {

  /**
   * Persist the new instance object into database. [create]
   * @param o transient object.
   * @throws IllegalArgumentException if given entity is already persistent.
   * @see org.hibernate.Session#persist(Object)
   */
  void create(@NotNull E o) throws DataAccessException, IllegalArgumentException;

  /**
   * Retrieve an object that was previously persisted to the database using the indicated id as primary key. [read]
   * @param id primary key
   * @return instance of object or <code>null</code> if object is not found.
   * @see org.hibernate.Session#get(Class, java.io.Serializable)
   */
  @Nullable
  E get(@NotNull K id) throws DataAccessException;

  /**
   * Update the given object. [update]
   * @param o instance of object to update.
   * @see org.hibernate.Session#update(Object)
   */
  void update(@NotNull E o) throws DataAccessException;

  /**
   * Remove an object from persistent storage in the database. [delete]
   * @param o object.
   * @see org.hibernate.Session#delete(Object)
   */
  void delete(@NotNull E o) throws DataAccessException;

  /**
   * Save or update object.
   * @param o object.
   * @see org.hibernate.Session#saveOrUpdate(Object)
   */
  void saveOrUpdate(@NotNull E o) throws DataAccessException;

  /**
   * Merge object with persistent object.
   * @param o object to merge.
   * @return merged object
   * @see org.hibernate.Session#merge(Object)
   */
  @NotNull
  E merge(@NotNull E o) throws DataAccessException;

  /**
   * Refresh the state of the instance from the database, overwriting changes made to the entity, if any.
   * @param o object.
   * @see org.hibernate.Session#refresh(Object)
   */
  void refresh(@NotNull final E o) throws DataAccessException;
}
