/**
 * AutoDAO - Generic DAO on steroids implementation for Java.
 * Copyright (C) 2008-2009  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.autodao;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.aopalliance.intercept.MethodInterceptor;
import org.aopalliance.intercept.MethodInvocation;
import org.jetbrains.annotations.NotNull;
import org.springframework.core.annotation.AnnotationUtils;
import org.springframework.transaction.interceptor.TransactionAspectSupport;

/**
 * Intercepts finder DAO calls and calls execute*Finder methods on dao.
 * <p/>
 * NOT FOR PUBLIC USE.
 */
public final class FinderIntroductionInterceptor extends TransactionAspectSupport implements MethodInterceptor {

  @NotNull
  private final QueryArgumentTransformer queryArgumentTransformer;

  public FinderIntroductionInterceptor(@NotNull final QueryArgumentTransformer queryArgumentTransformer) {
    this.queryArgumentTransformer = queryArgumentTransformer;
  }

  /** {@inheritDoc} */
  @Override
  public Object invoke(@NotNull final MethodInvocation invocation) throws Throwable {
    final Method method = invocation.getMethod();
    final Finder finder = AnnotationUtils.getAnnotation(method, Finder.class);

    if (finder == null) {
      return invocation.proceed();
    } else {
      final Object[] arguments = invocation.getArguments();
      final Integer[] limit = new Integer[1];
      final Integer[] offset = new Integer[1];
      final List<Object> queryArgs = new ArrayList<Object>(arguments.length);
      final Map<String, Object> namedArgs = new HashMap<String, Object>(arguments.length);
      Utils.visitMethodParameters(method, new Utils.ParameterCallback() {
        @Override
        public void visit(final int index, final Class<?> type, final Annotation[] annotations) {
          boolean processed = false;
          final Object arg = arguments[index];
          if (Utils.getLimit(annotations) != null) {
            limit[0] = (Integer) arg;
            processed = true;
          }
          if (Utils.getOffset(annotations) != null) {
            offset[0] = (Integer) arg;
            processed = true;
          }
          if (!processed) {
            final Object queryArg = queryArgumentTransformer.transform(type, annotations, arg);
            final Named named = Utils.getNamed(annotations);
            if (named == null) {
              queryArgs.add(queryArg);
            } else {
              namedArgs.put(named.value(), queryArg);
            }
          }
        }
      });

      final FinderExecutor dao = (FinderExecutor) invocation.getThis();
      if (Collection.class.isAssignableFrom(invocation.getMethod().getReturnType())) {
        final List<Object> data;
        if (namedArgs.isEmpty()) {
          data = dao.executeListFinder(method, queryArgs.toArray(new Object[queryArgs.size()]), limit[0], offset[0]);
        } else {
          data = dao.executeListFinder(method, namedArgs, limit[0], offset[0]);
        }
        return convertResult(finder, data);
      } else {
        if (namedArgs.isEmpty()) {
          return dao.executeSingleFinder(method, arguments);
        } else {
          return dao.executeSingleFinder(method, namedArgs);
        }
      }
    }
  }

  @SuppressWarnings({ "unchecked" })
  private static Object convertResult(final Finder finder, final List<Object> data) throws Exception {
    final Collection<Object> result;
    if (finder.returnAs() == List.class) {
      result = data;
    } else {
      result = (Collection<Object>) finder.returnAs().newInstance();
      result.addAll(data);
    }
    return result;
  }
}
