/**
 * AutoDAO - Generic DAO on steroids implementation for Java.
 * Copyright (C) 2008-2009  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.autodao.hibernate;

import java.lang.reflect.Method;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import net.sf.autodao.Finder;
import net.sf.autodao.FinderExecutor;
import net.sf.autodao.Utils;
import org.hibernate.Query;
import org.hibernate.Session;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.springframework.core.annotation.AnnotationUtils;
import org.springframework.orm.hibernate3.HibernateCallback;
import org.springframework.orm.hibernate3.HibernateTemplate;
import org.springframework.util.StringUtils;

/**
 * Hibernate implementation of AutoDAO.
 * <p/>
 * NOT FOR PUBLIC USE.
 */
class HibernateDao implements FinderExecutor {

  protected final HibernateTemplate hibernateTemplate;

  HibernateDao(@NotNull final HibernateTemplate hibernateTemplate) {
    this.hibernateTemplate = hibernateTemplate;
  }

  /** {@inheritDoc} */
  @NotNull
  @Override
  @SuppressWarnings({ "unchecked" })
  public final <T> List<T> executeListFinder(@NotNull final Method method,
                                             @NotNull final Map<String, ?> queryArgs,
                                             @Nullable final Integer limit,
                                             @Nullable final Integer offset) {
    return (List<T>) hibernateTemplate.executeFind(new HibernateCallback() {
      @Override
      public Object doInHibernate(final Session session) {
        return populateParams(buildQuery(session, method, limit, offset), queryArgs).list();
      }
    });
  }

  /** {@inheritDoc} */
  @NotNull
  @Override
  @SuppressWarnings({ "unchecked" })
  public final <T> List<T> executeListFinder(@NotNull final Method method,
                                             @NotNull final Object[] queryArgs,
                                             @Nullable final Integer limit,
                                             @Nullable final Integer offset) {
    return (List<T>) hibernateTemplate.executeFind(new HibernateCallback() {
      @Override
      public Object doInHibernate(final Session session) {
        return populateParams(buildQuery(session, method, limit, offset), queryArgs).list();
      }
    });
  }

  /** {@inheritDoc} */
  @Override
  @SuppressWarnings({ "unchecked" })
  public final <T> T executeSingleFinder(@NotNull final Method method, @NotNull final Map<String, ?> queryArgs) {
    return (T) hibernateTemplate.execute(new HibernateCallback() {
      @Override
      public Object doInHibernate(final Session session) {
        return populateParams(buildQuery(session, method, null, null), queryArgs).uniqueResult();
      }
    });
  }

  /** {@inheritDoc} */
  @Override
  @SuppressWarnings({ "unchecked" })
  public final <T> T executeSingleFinder(@NotNull final Method method, @NotNull final Object[] queryArgs) {
    return (T) hibernateTemplate.execute(new HibernateCallback() {
      @Override
      public Object doInHibernate(final Session session) {
        return populateParams(buildQuery(session, method, null, null), queryArgs).uniqueResult();
      }
    });
  }

  @NotNull
  private Query buildQuery(@NotNull final Session s,
                           @NotNull final Method m,
                           @Nullable final Integer limit,
                           @Nullable final Integer offset) {
    final Finder f = AnnotationUtils.getAnnotation(m, Finder.class);
    final Query q;
    if (StringUtils.hasText(f.query())) {
      q = s.createQuery(f.query());
    } else if (StringUtils.hasText(f.sqlQuery())) {
      q = s.createSQLQuery(f.sqlQuery());
    } else if (StringUtils.hasText(f.sqlQueryName())) {
      q = s.getNamedQuery(f.sqlQueryName());
    } else if (StringUtils.hasText(f.queryName())) {
      q = s.getNamedQuery(f.queryName());
    } else {
      throw new IllegalArgumentException("Don't know how to handle " + m + " method.\n"
          + "Please update to latest AutoDAO release.\n"
          + "If that doesn't help, please file a bugreport to AutoDAO bugtracker.");
    }
    if (limit != null) {
      q.setMaxResults(limit);
    }
    if (offset != null) {
      q.setFirstResult(offset);
    }
    return q;
  }

  @NotNull
  private Query populateParams(@NotNull final Query q, @NotNull final Map<String, ?> params) {
    for (final Map.Entry<String, ?> o : params.entrySet()) {
      final String paramName = o.getKey();
      final Object value = o.getValue();
      if (value == null) {
        q.setParameter(paramName, null);
      } else if (value.getClass().isArray()) {
        q.setParameterList(paramName, Utils.toObjectArray(value));
      } else if (Collection.class.isAssignableFrom(value.getClass())) {
        q.setParameterList(paramName, (Collection) value);
      } else {
        q.setParameter(paramName, value);
      }
    }
    return q;
  }

  @NotNull
  private Query populateParams(@NotNull final Query q, @NotNull final Object[] params) {
    int i = 0;
    for (final Object o : params) {
      q.setParameter(i, o);
      i++;
    }
    return q;
  }
}
