/**
 * AutoDAO - Generic DAO on steroids implementation for Java.
 * Copyright (C) 2008-2009  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.autodao.hibernate;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.sf.autodao.Dao;
import net.sf.autodao.DefaultQueryArgumentTransformer;
import net.sf.autodao.Finder;
import net.sf.autodao.FinderIntroductionInterceptor;
import net.sf.autodao.Named;
import net.sf.autodao.PersistentEntity;
import net.sf.autodao.QueryArgumentTransformer;
import net.sf.autodao.SpecificationViolationException;
import net.sf.autodao.Utils;
import org.hibernate.QueryParameterException;
import org.hibernate.SessionFactory;
import org.hibernate.engine.SessionFactoryImplementor;
import org.hibernate.engine.query.NamedParameterDescriptor;
import org.hibernate.engine.query.OrdinalParameterDescriptor;
import org.hibernate.engine.query.ParameterMetadata;
import org.hibernate.engine.query.QueryPlanCache;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.aop.framework.ProxyFactoryBean;
import org.springframework.beans.factory.BeanFactory;
import org.springframework.beans.factory.NoSuchBeanDefinitionException;
import org.springframework.core.annotation.AnnotationUtils;
import org.springframework.orm.hibernate3.HibernateTemplate;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.TransactionDefinition;
import org.springframework.transaction.interceptor.DefaultTransactionAttribute;
import org.springframework.transaction.interceptor.MatchAlwaysTransactionAttributeSource;
import org.springframework.transaction.interceptor.TransactionInterceptor;
import org.springframework.util.Assert;
import static org.springframework.util.StringUtils.hasText;

/**
 * Validating proxy factory.
 * <p/>
 * NOT FOR PUBLIC USE.
 */
final class HibernateDaoProxyFactoryBean extends ProxyFactoryBean {
  private static final String QUERY_ARGUMENT_TRANSFORMER_BEAN_NAME = "autodaoQueryArgumentTransformer";

  /** Log. Just log. */
  private static final Logger LOG = LoggerFactory.getLogger(HibernateDaoProxyFactoryBean.class);

  public HibernateDaoProxyFactoryBean(@NotNull final Class<?> daoInterface,
                                      @NotNull final SessionFactory sessionFactory,
                                      @NotNull final PlatformTransactionManager transactionManager,
                                      @Nullable final Class<? extends PersistentEntity<?>> entityType) {
    addInterface(daoInterface);
    validate(daoInterface, sessionFactory);
    @Nullable
    final Class<? extends PersistentEntity<?>> realEntityType = getEntityType(daoInterface, entityType);
    addAdvice(createTransactionInterceptor(transactionManager));
    setTarget(createDao(daoInterface, realEntityType, sessionFactory));
  }

  @NotNull
  private static TransactionInterceptor createTransactionInterceptor(@NotNull final PlatformTransactionManager transactionManager) {
    final MatchAlwaysTransactionAttributeSource source = new MatchAlwaysTransactionAttributeSource();
    source.setTransactionAttribute(new DefaultTransactionAttribute(TransactionDefinition.PROPAGATION_MANDATORY));
    return new TransactionInterceptor(transactionManager, source);
  }

  @Override
  public void setBeanFactory(final BeanFactory beanFactory) {
    QueryArgumentTransformer queryArgumentTransformer;
    try {
      queryArgumentTransformer = (QueryArgumentTransformer) beanFactory.getBean(QUERY_ARGUMENT_TRANSFORMER_BEAN_NAME, QueryArgumentTransformer.class);
    } catch (final NoSuchBeanDefinitionException ignore) {
      queryArgumentTransformer = new DefaultQueryArgumentTransformer();
    }
    addAdvice(new FinderIntroductionInterceptor(queryArgumentTransformer));
    super.setBeanFactory(beanFactory);
  }

  private static Class<? extends PersistentEntity<?>> getEntityType(@NotNull final Class<?> daoInterface,
                                                                    @Nullable final Class<? extends PersistentEntity<?>> entityType) {
    final Class<? extends PersistentEntity<?>> realEntityType;
    if (Dao.class.isAssignableFrom(daoInterface)) {
      if (entityType == null) {
        realEntityType = autoDiscoverEntityType(daoInterface);
      } else {
        realEntityType = entityType;
      }
      Assert.notNull(realEntityType, "No entity type found for " + daoInterface);
      //See NoEntityTest
      Assert.isAssignable(PersistentEntity.class, realEntityType, "Invalid entity type: ");
    } else {
      if (entityType != null) {
        LOG.warn("Useless entity type for dao " + daoInterface + ", skipping...");
      }
      realEntityType = null;
    }
    return realEntityType;
  }

  @SuppressWarnings({ "unchecked" })
  @NotNull
  private static Object createDao(@NotNull final Class<?> daoInterface,
                                  @Nullable final Class<?> entityType,
                                  @NotNull final SessionFactory sessionFactory) {
    final HibernateTemplate hibernateTemplate = new HibernateTemplate(sessionFactory, false);
    if (Dao.class.isAssignableFrom(daoInterface)) {
      assert entityType != null;
      final String msg = entityType.getName() + " is not a persistent class (it is not registered in SessionFactory)";
      Assert.notNull(sessionFactory.getClassMetadata(entityType), msg);
      return new HibernateEntityDao(entityType, hibernateTemplate);
    } else {
      return new HibernateDao(hibernateTemplate);
    }
  }

  @SuppressWarnings({ "unchecked" })
  private static Class<? extends PersistentEntity<?>> autoDiscoverEntityType(final Class<?> daoInterface) {
    //Automagically discover entity type from generics.
    for (final Type t : daoInterface.getGenericInterfaces()) {
      if (t instanceof ParameterizedType) {
        final ParameterizedType pt = (ParameterizedType) t;
        if (pt.getRawType() == Dao.class) {
          final Type type = pt.getActualTypeArguments()[0];
          if (type instanceof Class) {
            return (Class<? extends PersistentEntity<?>>) type;
          } else if (type instanceof ParameterizedType) {
            //Special handling for cases like Dao<Entity<Foo>>.
            //See ParamEntityTest
            return (Class<? extends PersistentEntity<?>>) ((ParameterizedType) type).getRawType();
          } else {
            throw new IllegalArgumentException("Don't know how to handle " + type + " entity type for " + daoInterface.getName() + ".\n"
                + "Please update to latest AutoDAO release.\n"
                + "If that doesn't help, please file a bugreport to AutoDAO bugtracker.");
          }
        }
      }
    }
    return null;
  }

  /**
   * Validates given DAO interface.
   * @param daoInterface   dao interface.
   * @param sessionFactory Hibernate SessionFactory.
   */
  private static void validate(final Class<?> daoInterface, final SessionFactory sessionFactory) {
    final SessionFactoryImplementor sfi;
    if (sessionFactory instanceof SessionFactoryImplementor) {
      sfi = (SessionFactoryImplementor) sessionFactory;
    } else {
      sfi = null;
      LOG.warn("Could not cast session factory " + sessionFactory + " to "
          + SessionFactoryImplementor.class.getName() + ". Some validations were disabled for DAO "
          + daoInterface.getName());
    }

    for (final Method m : daoInterface.getMethods()) {
      final Finder f = AnnotationUtils.getAnnotation(m, Finder.class);
      if (f == null) {
        if (!Arrays.asList(Dao.class.getMethods()).contains(m)) {
          throw new SpecificationViolationException("Method is neither a finder method nor is implemented in AutoDAO", m);
        }
      } else {
        final boolean singleFind = validateReturnType(m, f);
        @Nullable
        final ParameterMetadata parameterMetadata = validateQuery(sfi, m, f);
        validateParams(m, singleFind, parameterMetadata);
      }
    }
  }

  private static void validateParams(@NotNull final Method m,
                                     final boolean singleFind,
                                     @Nullable
                                     final ParameterMetadata metadata) {
    final boolean[] limit = new boolean[1];
    final boolean[] offset = new boolean[1];
    final int[] indexed = new int[1];
    final Set<String> names = new HashSet<String>();

    Utils.visitMethodParameters(m, new Utils.ParameterCallback() {
      @Override
      public void visit(final int index, final Class<?> type, final Annotation[] annotations) {
        boolean processed = false;
        if (Utils.getLimit(annotations) != null) {
          limitOffsetCheck(limit, singleFind, type, m);
          processed = true;
        }
        if (Utils.getOffset(annotations) != null) {
          limitOffsetCheck(offset, singleFind, type, m);
          processed = true;
        }
        final Named named = Utils.getNamed(annotations);
        if (processed) {
          if (named != null) {
            throw new SpecificationViolationException("@Named cannot be present on @Limit/@Offset param", m);
          }
        } else {
          if (named == null) {
            if (!names.isEmpty()) {
              throw new SpecificationViolationException("You're not allowed to mix @Named and indexed params", m);
            }
            if (metadata != null) {
              final int paramIndex = index + 1
                  - (limit[0] ? 1 : 0)
                  - (offset[0] ? 1 : 0);
              final OrdinalParameterDescriptor descriptor;
              try {
                descriptor = metadata.getOrdinalParameterDescriptor(paramIndex);
              } catch (final IndexOutOfBoundsException e) {
                throw new SpecificationViolationException("Query doesn't have enough indexed parameters", m, e);
              }
              /*final Class<?> expectedType = descriptor.getExpectedType().getReturnedClass();
              if (!expectedType.isAssignableFrom(type)) {
                final String msg = String.format("Indexed parameter '%s' of type %s cannot be assigned to query parameter of type %s", paramIndex, type.getName(), expectedType.getName());
                throw new SpecificationViolationException(msg, m);
              }*/
            }
            indexed[0]++;
          } else {
            if (indexed[0] > 0) {
              throw new SpecificationViolationException("You're not allowed to mix @Named and indexed params", m);
            }
            final String name = named.value();
            if (metadata != null) {
              final NamedParameterDescriptor descriptor;
              try {
                descriptor = metadata.getNamedParameterDescriptor(name);
              } catch (final QueryParameterException e) {
                throw new SpecificationViolationException(String.format("Query doesn't have named parameter '%s'", name), m, e);
              }
              /*final Class<?> expectedType = descriptor.getExpectedType().getReturnedClass();
              if (!expectedType.isAssignableFrom(type)) {
                final String msg = String.format("Named parameter '%s' of type %s cannot be assigned to query parameter of type %s", name, type.getName(), expectedType.getName());
                throw new SpecificationViolationException(msg, m);
              }*/
            }
            if (!names.add(name)) {
              throw new SpecificationViolationException("Duplicate named parameters '" + name + "'", m);
            }
          }
        }
      }
    });
    if (metadata != null) {
      if (indexed[0] != metadata.getOrdinalParameterCount()) {
        throw new SpecificationViolationException(String.format("Not enough indexed params (query has %s but method has %s)", metadata.getOrdinalParameterCount(), indexed[0]), m);
      } else if (!names.equals(metadata.getNamedParameterNames())) {
        throw new SpecificationViolationException(String.format("Wrong named parameters (query has %s but method has %s)", metadata.getNamedParameterNames(), names), m);
      }
    }
  }

  private static void limitOffsetCheck(final boolean[] holder,
                                       final boolean singleFind,
                                       final Class<?> type,
                                       final Method m) {
    if (holder[0]) {
      throw new SpecificationViolationException("@Limit/@Offset can appear only on one parameter", m);
    } else if (singleFind) {
      throw new SpecificationViolationException("@Limit/@Offset isn't allowed on single-object finders", m);
    }
    if (type != Integer.class && type != Integer.TYPE) {
      throw new SpecificationViolationException(
          "@Limit and @Offset argument type must be either int or java.lang.Integer", m);
    }
    holder[0] = true;
  }

  private static boolean validateReturnType(final Method m, final Finder f) {
    final boolean singleFind = !Collection.class.isAssignableFrom(m.getReturnType());
    if (!singleFind) {
      final Class<?> returnType = m.getReturnType();
      if (!returnType.isAssignableFrom(f.returnAs())) {
        throw new SpecificationViolationException(
            "@Finder.returnAs value (" + f.returnAs().getName()
                + ") cannot be cast to method return type (" + returnType.getName() + ")", m
        );
      } else if (f.returnAs() != List.class) {
        if (f.returnAs().isInterface()) {
          throw new SpecificationViolationException(
              "@Finder.returnAs value (" + f.returnAs().getName() + ") cannot be an interface", m
          );
        } else {
          try {
            f.returnAs().getConstructor();
          } catch (NoSuchMethodException e) {
            throw new SpecificationViolationException(
                "@Finder.returnAs value (" + f.returnAs().getName() + ") doesn't have no-arg constructor.", m
            );
          }
        }
      }
    }
    return singleFind;
  }

  @Nullable
  private static ParameterMetadata validateQuery(@Nullable final SessionFactoryImplementor sfi,
                                                 @NotNull final Method m,
                                                 @NotNull final Finder f) {
    final String query = f.query();
    final String queryName = f.queryName();
    final String sqlQuery = f.sqlQuery();
    final String sqlQueryName = f.sqlQueryName();
    final int sum
        = (hasText(query) ? 1 : 0)
        + (hasText(queryName) ? 1 : 0)
        + (hasText(sqlQuery) ? 1 : 0)
        + (hasText(sqlQueryName) ? 1 : 0);
    if (sum < 1) {
      throw new SpecificationViolationException("You need to specify at least one of: query, queryName, sqlQuery, sqlQueryName", m);
    } else if (sum > 1) {
      throw new SpecificationViolationException("You can specify only one of: query, queryName, sqlQuery, sqlQueryName", m);
    }
    if (sfi != null) {
      final QueryPlanCache cache = sfi.getQueryPlanCache();
      if (hasText(query)) {
        try {
          return cache.getHQLQueryPlan(query, false, Collections.emptyMap()).getParameterMetadata();
        } catch (final Exception e) {
          throw new SpecificationViolationException("Invalid query: " + query, m, e);
        }
      } else if (hasText(sqlQuery)) {
        return cache.getSQLParameterMetadata(sqlQuery);
      } else if (hasText(sqlQueryName)) {
        if (sfi.getNamedSQLQuery(sqlQueryName) == null) {
          throw new SpecificationViolationException("No named SQL query '" + sqlQueryName + "' found.", m);
        }
        return cache.getSQLParameterMetadata(sfi.getNamedSQLQuery(sqlQueryName).getQuery());
      } else if (hasText(queryName)) {
        if (sfi.getNamedQuery(queryName) == null) {
          throw new SpecificationViolationException("No named query '" + queryName + "' found.", m);
        }
        return cache.getHQLQueryPlan(sfi.getNamedQuery(queryName).getQuery(), false, Collections.emptyMap()).getParameterMetadata();
      } else {
        throw new IllegalArgumentException("Don't know how to handle " + m + ".\n"
            + "Please update to latest AutoDAO release.\n"
            + "If that doesn't help, please file a bugreport to AutoDAO bugtracker.");
      }
    } else {
      return null;
    }
  }
}
