/**
 * AutoDAO - Generic DAO on steroids implementation for Java.
 * Copyright (C) 2008-2009  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.autodao.hibernate;

import java.io.Serializable;

import net.sf.autodao.Dao;
import net.sf.autodao.PersistentEntity;
import org.jetbrains.annotations.NotNull;
import org.springframework.orm.hibernate3.HibernateTemplate;
import org.springframework.util.Assert;

/**
 * Hibernate implementation of Entity AutoDAO.
 * <p/>
 * NOT FOR PUBLIC USE.
 */
final class HibernateEntityDao<E extends PersistentEntity<K>, K extends Serializable>
    extends HibernateDao implements Dao<E, K> {


  /** class of persistent object. */
  @NotNull
  private final Class<E> entityType;

  HibernateEntityDao(@NotNull final Class<E> entityType, @NotNull final HibernateTemplate hibernateTemplate) {
    super(hibernateTemplate);
    this.entityType = entityType;
  }

  /** {@inheritDoc} */
  @Override
  public void create(@NotNull final E o) {
    checkObject(o);
    Assert.isTrue(!hibernateTemplate.contains(o), "You're trying to persist already persistent entity: " + o);
    hibernateTemplate.persist(o);
  }

  /** {@inheritDoc} */
  @Override
  public void delete(@NotNull final E o) {
    checkObject(o);
    Assert.isTrue(hibernateTemplate.contains(o), "You're trying to delete transient/detached entity: " + o);
    hibernateTemplate.delete(o);
  }

  /** {@inheritDoc} */
  @Override
  @SuppressWarnings({ "unchecked" })
  public E get(@NotNull final K id) {
    Assert.notNull(id, "Id cannot be null");
    return (E) hibernateTemplate.get(entityType, id);
  }

  /** {@inheritDoc} */
  @NotNull
  @Override
  @SuppressWarnings({ "unchecked" })
  public E merge(@NotNull final E o) {
    return (E) hibernateTemplate.merge(o);
  }

  /** {@inheritDoc} */
  @Override
  public void refresh(@NotNull final E o) {
    hibernateTemplate.refresh(o);
  }

  /** {@inheritDoc} */
  @Override
  public void saveOrUpdate(@NotNull final E o) {
    checkObject(o);
    hibernateTemplate.saveOrUpdate(o);
  }

  /** {@inheritDoc} */
  @Override
  public String toString() {
    return "Entity Dao for " + entityType.getName();
  }

  /** {@inheritDoc} */
  @Override
  public void update(@NotNull final E o) {
    checkObject(o);
    Assert.isTrue(hibernateTemplate.contains(o), "You're trying to update transient/detached entity: " + o);
    hibernateTemplate.update(o);
  }

  /**
   * Check entity.
   * @param o entity.
   */
  private void checkObject(@NotNull final E o) {
    Assert.notNull(o, "Entity cannot be null");
    Assert.isAssignable(entityType, o.getClass(), "Invalid entity type");
  }
}
